use strict;
use warnings;
use Groonga::API::Test;

plan skip_all => 'requires groonga version >= 2.0.9' unless version_ge("2.0.9");

db_test(sub {
  my ($ctx, $db) = @_;

  my $path = "./groonga_query.log";
  if (version_ge("2.1.2")) {
    unlink $path if -f $path;
    Groonga::API::default_query_logger_set_path($path);
    is Groonga::API::default_query_logger_get_path() => $path, "correct path";
  }

  Groonga::API::default_query_logger_set_flags(GRN_QUERY_LOG_ALL);

  my $rc = Groonga::API::query_logger_pass($ctx, GRN_QUERY_LOG_CACHE);
  ok $rc, "should log CACHE";

  $rc = Groonga::API::query_logger_pass($ctx, GRN_QUERY_LOG_COMMAND);
  ok $rc, "should also log COMMAND";

  Groonga::API::default_query_logger_set_flags(GRN_QUERY_LOG_COMMAND);

  $rc = Groonga::API::query_logger_pass($ctx, GRN_QUERY_LOG_SIZE);
  ok !$rc, "should not log SIZE now";

  $rc = Groonga::API::query_logger_pass($ctx, GRN_QUERY_LOG_COMMAND);
  ok $rc, "should still log COMMAND";

  Groonga::API::query_logger_put($ctx, GRN_QUERY_LOG_COMMAND, 'test', "command");
  Groonga::API::query_logger_put($ctx, GRN_QUERY_LOG_COMMAND, 'test', '%s', "command");
  Groonga::API::query_logger_put($ctx, GRN_QUERY_LOG_COMMAND, 'test', '%s %s', "command", "command");

  if (version_ge("2.1.2")) {
    ok -s $path, "log file has been written";
    note do { open my $fh, '<', $path; local $/; <$fh> };
    unlink $path if -f $path;
  }

  if (version_ge("2.1.0")) {
    Groonga::API::query_logger_reopen($ctx); # 2.1.0

    Groonga::API::query_logger_put($ctx, GRN_QUERY_LOG_COMMAND, 'test', "command");
    Groonga::API::query_logger_put($ctx, GRN_QUERY_LOG_COMMAND, 'test', '%s', "command");
    Groonga::API::query_logger_put($ctx, GRN_QUERY_LOG_COMMAND, 'test', '%s %s', "command", "command");

    if (version_ge("2.1.2")) {
      ok -s $path, "log file has been written";
      note do { open my $fh, '<', $path; local $/; <$fh> };
      unlink $path if -f $path;
    }
  }
});

# TODO: GRN_LOG() support?

done_testing;
